@@ -11,6 +11,7 @@
 #include <libknot/rrtype/rdname.h>
 #include <libknot/descriptor.h>
 #include <ucw/mempool.h>
+#include <sys/socket.h>
 #include "kresconfig.h"
 #include "lib/resolve.h"
 #include "lib/layer.h"
@@ -110,6 +111,12 @@ static int answer_finalize_yield(kr_layer_t *ctx) { return kr_ok(); }
 			struct kr_layer layer = {.state = (r)->state, .api = mod->layer, .req = (r)}; \
 			if (layer.api && layer.api->func) { \
 				(r)->state = layer.api->func(&layer, ##__VA_ARGS__); \
+				/* It's an easy mistake to return error code, for example. */ \
+				/* (though we could allow such an overload later) */ \
+				if (unlikely(!kr_state_consistent((r)->state))) { \
+					assert(!EINVAL); \
+					(r)->state = KR_STATE_FAIL; \
+				} else \
 				if ((r)->state == KR_STATE_YIELD) { \
 					func ## _yield(&layer, ##__VA_ARGS__); \
 					break; \
@@ -147,7 +154,7 @@ static void randomized_qname_case(knot_dname_t * restrict qname, uint32_t secret
 		return;
 	}
 	assert(qname);
-	const int len = knot_dname_size(qname) - 2; /* Skip first, last label. */
+	const int len = knot_dname_size(qname) - 2; /* Skip first, last label. First is length, last is always root */
 	for (int i = 0; i < len; ++i) {
 		/* Note: this relies on the fact that correct label lengths
 		 * can't pass the isletter() test (by "luck"). */
@@ -157,23 +164,6 @@ static void randomized_qname_case(knot_dname_t * restrict qname, uint32_t secret
 	}
 }
 
-/** Invalidate current NS/addr pair. */
-static int invalidate_ns(struct kr_rplan *rplan, struct kr_query *qry)
-{
-	if (qry->ns.addr[0].ip.sa_family != AF_UNSPEC) {
-		const char *addr = kr_inaddr(&qry->ns.addr[0].ip);
-		int addr_len = kr_inaddr_len(&qry->ns.addr[0].ip);
-		int ret = kr_zonecut_del(&qry->zone_cut, qry->ns.name, addr, addr_len);
-		/* Also remove it from the qry->ns.addr array.
-		 * That's useful at least for STUB and FORWARD modes. */
-		memmove(qry->ns.addr, qry->ns.addr + 1,
-			sizeof(qry->ns.addr[0]) * (KR_NSREP_MAXADDR - 1));
-		return ret;
-	} else {
-		return kr_zonecut_del_all(&qry->zone_cut, qry->ns.name);
-	}
-}
-
 /** This turns of QNAME minimisation if there is a non-terminal between current zone cut, and name target.
  *  It save several minimization steps, as the zone cut is likely final one.
  */
@@ -310,71 +300,6 @@ static int ns_fetch_cut(struct kr_query *qry, const knot_dname_t *requested_name
 	return KR_STATE_PRODUCE;
 }
 
-static int ns_resolve_addr(struct kr_query *qry, struct kr_request *req)
-{
-	struct kr_rplan *rplan = &req->rplan;
-	struct kr_context *ctx = req->ctx;
-
-
-	/* Start NS queries from root, to avoid certain cases
-	 * where a NS drops out of cache and the rest is unavailable,
-	 * this would lead to dependency loop in current zone cut.
-	 * Prefer IPv6 and continue with IPv4 if not available.
-	 */
-	uint16_t next_type = 0;
-	if (!(qry->flags.AWAIT_IPV6) &&
-	    !(ctx->options.NO_IPV6)) {
-		next_type = KNOT_RRTYPE_AAAA;
-		qry->flags.AWAIT_IPV6 = true;
-	} else if (!(qry->flags.AWAIT_IPV4) &&
-		   !(ctx->options.NO_IPV4)) {
-		next_type = KNOT_RRTYPE_A;
-		qry->flags.AWAIT_IPV4 = true;
-		/* Hmm, no useable IPv6 then. */
-		qry->ns.reputation |= KR_NS_NOIP6;
-		kr_nsrep_update_rep(&qry->ns, qry->ns.reputation, ctx->cache_rep);
-	}
-	/* Bail out if the query is already pending or dependency loop. */
-	if (!next_type || kr_rplan_satisfies(qry->parent, qry->ns.name, KNOT_CLASS_IN, next_type)) {
-		/* Fall back to SBELT if root server query fails. */
-		if (!next_type && qry->zone_cut.name[0] == '\0') {
-			VERBOSE_MSG(qry, "=> fallback to root hints\n");
-			kr_zonecut_set_sbelt(ctx, &qry->zone_cut);
-			qry->flags.NO_THROTTLE = true; /* Pick even bad SBELT servers */
-			return kr_error(EAGAIN);
-		}
-		/* No IPv4 nor IPv6, flag server as unusable. */
-		++req->count_no_nsaddr;
-		VERBOSE_MSG(qry, "=> unresolvable NS address, bailing out (counter: %u)\n",
-				req->count_no_nsaddr);
-		qry->ns.reputation |= KR_NS_NOIP4 | KR_NS_NOIP6;
-		kr_nsrep_update_rep(&qry->ns, qry->ns.reputation, ctx->cache_rep);
-		invalidate_ns(rplan, qry);
-		return kr_error(EHOSTUNREACH);
-	}
-	/* Push new query to the resolution plan */
-	struct kr_query *next =
-		kr_rplan_push(rplan, qry, qry->ns.name, KNOT_CLASS_IN, next_type);
-	if (!next) {
-		return kr_error(ENOMEM);
-	}
-	next->flags.NONAUTH = true;
-
-	/* At the root level with no NS addresses, add SBELT subrequest. */
-	int ret = 0;
-	if (qry->zone_cut.name[0] == '\0') {
-		ret = kr_zonecut_set_sbelt(ctx, &next->zone_cut);
-		if (ret == 0) { /* Copy TA and key since it's the same cut to avoid lookup. */
-			kr_zonecut_copy_trust(&next->zone_cut, &qry->zone_cut);
-			kr_zonecut_set_sbelt(ctx, &qry->zone_cut); /* Add SBELT to parent in case query fails. */
-			qry->flags.NO_THROTTLE = true; /* Pick even bad SBELT servers */
-		}
-	} else {
-		next->flags.AWAIT_CUT = true;
-	}
-	return ret;
-}
-
 static int edns_put(knot_pkt_t *pkt, bool reclaim)
 {
 	if (!pkt->opt_rr) {
@@ -681,7 +606,7 @@ static void answer_finalize(struct kr_request *request)
 static int query_finalize(struct kr_request *request, struct kr_query *qry, knot_pkt_t *pkt)
 {
 	knot_pkt_begin(pkt, KNOT_ADDITIONAL);
-	if (qry->flags.SAFEMODE)
+	if (qry->flags.NO_EDNS)
 		return kr_ok();
 	/* Remove any EDNS records from any previous iteration. */
 	int ret = edns_erase_and_reserve(pkt);
@@ -776,7 +701,7 @@ static int resolve_query(struct kr_request *request, const knot_pkt_t *packet)
 	return request->state;
 }
 
-knot_pkt_t * kr_request_ensure_answer(struct kr_request *request)
+knot_pkt_t *kr_request_ensure_answer(struct kr_request *request)
 {
 	if (request->answer)
 		return request->answer;
@@ -839,84 +764,6 @@ enomem:
 	return request->answer = NULL;
 }
 
-KR_PURE static bool kr_inaddr_equal(const struct sockaddr *a, const struct sockaddr *b)
-{
-	const int a_len = kr_inaddr_len(a);
-	const int b_len = kr_inaddr_len(b);
-	return a_len == b_len && memcmp(kr_inaddr(a), kr_inaddr(b), a_len) == 0;
-}
-
-static void update_nslist_rtt(struct kr_context *ctx, struct kr_query *qry, const struct sockaddr *src)
-{
-	/* Do not track in safe mode. */
-	if (qry->flags.SAFEMODE) {
-		return;
-	}
-
-	/* Calculate total resolution time from the time the query was generated. */
-	uint64_t elapsed = kr_now() - qry->timestamp_mono;
-	elapsed = elapsed > UINT_MAX ? UINT_MAX : elapsed;
-
-	/* NSs in the preference list prior to the one who responded will be penalised
-	 * with the RETRY timer interval. This is because we know they didn't respond
-	 * for N retries, so their RTT must be at least N * RETRY.
-	 * The NS in the preference list that responded will have RTT relative to the
-	 * time when the query was sent out, not when it was originated.
-	 */
-	for (size_t i = 0; i < KR_NSREP_MAXADDR; ++i) {
-		const struct sockaddr *addr = &qry->ns.addr[i].ip;
-		if (addr->sa_family == AF_UNSPEC) {
-			break;
-		}
-		/* If this address is the source of the answer, update its RTT */
-		if (kr_inaddr_equal(src, addr)) {
-			kr_nsrep_update_rtt(&qry->ns, addr, elapsed, ctx->cache_rtt, KR_NS_UPDATE);
-			WITH_VERBOSE(qry) {
-				char addr_str[INET6_ADDRSTRLEN];
-				inet_ntop(addr->sa_family, kr_inaddr(addr), addr_str, sizeof(addr_str));
-				VERBOSE_MSG(qry, "<= server: '%s' rtt: %"PRIu64" ms\n",
-						addr_str, elapsed);
-			}
-		} else {
-			/* Response didn't come from this IP, but we know the RTT must be at least
-			 * several RETRY timer tries, e.g. if we have addresses [a, b, c] and we have
-			 * tried [a, b] when the answer from 'a' came after 350ms, then we know
-			 * that 'b' didn't respond for at least 350 - (1 * 300) ms. We can't say that
-			 * its RTT is 50ms, but we can say that its score shouldn't be less than 50. */
-			 kr_nsrep_update_rtt(&qry->ns, addr, elapsed, ctx->cache_rtt, KR_NS_MAX);
-			 WITH_VERBOSE(qry) {
-			 	char addr_str[INET6_ADDRSTRLEN];
-			 	inet_ntop(addr->sa_family, kr_inaddr(addr), addr_str, sizeof(addr_str));
-				VERBOSE_MSG(qry, "<= server: '%s' rtt: >= %"PRIu64" ms\n",
-						addr_str, elapsed);
-			 }
-		}
-		/* Subtract query start time from elapsed time */
-		if (elapsed < KR_CONN_RETRY) {
-			break;
-		}
-		elapsed = elapsed - KR_CONN_RETRY;
-	}
-}
-
-static void update_nslist_score(struct kr_request *request, struct kr_query *qry, const struct sockaddr *src, knot_pkt_t *packet)
-{
-	struct kr_context *ctx = request->ctx;
-	/* On successful answer, update preference list RTT and penalise timer  */
-	if (!(request->state & KR_STATE_FAIL)) {
-		/* Update RTT information for preference list */
-		update_nslist_rtt(ctx, qry, src);
-		/* Do not complete NS address resolution on soft-fail. */
-		const int rcode = packet ? knot_wire_get_rcode(packet->wire) : 0;
-		if (rcode != KNOT_RCODE_SERVFAIL && rcode != KNOT_RCODE_REFUSED) {
-			qry->flags.AWAIT_IPV6 = false;
-			qry->flags.AWAIT_IPV4 = false;
-		} else { /* Penalize SERVFAILs. */
-			kr_nsrep_update_rtt(&qry->ns, src, KR_NS_PENALTY, ctx->cache_rtt, KR_NS_ADD);
-		}
-	}
-}
-
 static bool resolution_time_exceeded(struct kr_query *qry, uint64_t now)
 {
 	uint64_t resolving_time = now - qry->creation_time_mono;
@@ -929,7 +776,7 @@ static bool resolution_time_exceeded(struct kr_query *qry, uint64_t now)
 	return false;
 }
 
-int kr_resolve_consume(struct kr_request *request, const struct sockaddr *src, knot_pkt_t *packet)
+int kr_resolve_consume(struct kr_request *request, struct kr_transport **transport, knot_pkt_t *packet)
 {
 	struct kr_rplan *rplan = &request->rplan;
 
@@ -946,11 +793,7 @@ int kr_resolve_consume(struct kr_request *request, const struct sockaddr *src, k
 	}
 	bool tried_tcp = (qry->flags.TCP);
 	if (!packet || packet->size == 0) {
-		if (tried_tcp) {
-			request->state = KR_STATE_FAIL;
-		} else {
-			qry->flags.TCP = true;
-		}
+		return KR_STATE_PRODUCE;
 	} else {
 		/* Packet cleared, derandomize QNAME. */
 		knot_dname_t *qname_raw = knot_pkt_qname(packet);
@@ -963,25 +806,29 @@ int kr_resolve_consume(struct kr_request *request, const struct sockaddr *src, k
 		} else {
 			/* Fill in source and latency information. */
 			request->upstream.rtt = kr_now() - qry->timestamp_mono;
-			request->upstream.addr = src;
+			request->upstream.transport = transport ? *transport : NULL;
 			ITERATE_LAYERS(request, qry, consume, packet);
 			/* Clear temporary information */
-			request->upstream.addr = NULL;
+			request->upstream.transport = NULL;
 			request->upstream.rtt = 0;
 		}
 	}
 
-	/* Track RTT for iterative answers */
-	if (src && !(qry->flags.CACHED)) {
-		update_nslist_score(request, qry, src, packet);
+	if (transport && !qry->flags.CACHED) {
+		if (!(request->state & KR_STATE_FAIL)) {
+			/* Do not complete NS address resolution on soft-fail. */
+			const int rcode = packet ? knot_wire_get_rcode(packet->wire) : 0;
+			if (rcode != KNOT_RCODE_SERVFAIL && rcode != KNOT_RCODE_REFUSED) {
+				qry->flags.AWAIT_IPV6 = false;
+				qry->flags.AWAIT_IPV4 = false;
+			}
+		}
 	}
-	/* Resolution failed, invalidate current NS. */
+
 	if (request->state & KR_STATE_FAIL) {
-		invalidate_ns(rplan, qry);
 		qry->flags.RESOLVED = false;
 	}
 
-	/* For multiple errors in a row; invalidate_ns() is not enough. */
 	if (!qry->flags.CACHED) {
 		if (request->state & KR_STATE_FAIL) {
 			if (++request->count_fail_row > KR_CONSUME_FAIL_ROW_LIMIT) {
@@ -991,6 +838,10 @@ int kr_resolve_consume(struct kr_request *request, const struct sockaddr *src, k
 						"bail out (mitigation for NXNSAttack "
 						"CVE-2020-12667)\n");
 				}
+				if (!qry->flags.NO_NS_FOUND) {
+					qry->flags.NO_NS_FOUND = true;
+					return KR_STATE_PRODUCE;
+				}
 				return KR_STATE_FAIL;
 			}
 		} else {
@@ -1016,7 +867,12 @@ int kr_resolve_consume(struct kr_request *request, const struct sockaddr *src, k
 
 	/* Do not finish with bogus answer. */
 	if (qry->flags.DNSSEC_BOGUS)  {
-		return KR_STATE_FAIL;
+		if (qry->flags.FORWARD || qry->flags.STUB) {
+			return KR_STATE_FAIL;
+		}
+		/* Other servers might not have broken DNSSEC. */
+		qry->flags.DNSSEC_BOGUS = false;
+		return KR_STATE_PRODUCE;
 	}
 
 	return kr_rplan_empty(&request->rplan) ? KR_STATE_DONE : KR_STATE_PRODUCE;
@@ -1368,17 +1224,79 @@ static int zone_cut_check(struct kr_request *request, struct kr_query *qry, knot
 	return trust_chain_check(request, qry);
 }
 
-int kr_resolve_produce(struct kr_request *request, struct sockaddr **dst, int *type, knot_pkt_t *packet)
+
+static int ns_resolve_addr(struct kr_query *qry, struct kr_request *param, struct kr_transport *transport, uint16_t next_type)
+{
+	struct kr_rplan *rplan = &param->rplan;
+	struct kr_context *ctx = param->ctx;
+
+
+	/* Start NS queries from root, to avoid certain cases
+	 * where a NS drops out of cache and the rest is unavailable,
+	 * this would lead to dependency loop in current zone cut.
+	 */
+
+	/* Bail out if the query is already pending or dependency loop. */
+	if (!next_type || kr_rplan_satisfies(qry->parent, transport->ns_name, KNOT_CLASS_IN, next_type)) {
+		/* Fall back to SBELT if root server query fails. */
+		if (!next_type && qry->zone_cut.name[0] == '\0') {
+			VERBOSE_MSG(qry, "=> fallback to root hints\n");
+			kr_zonecut_set_sbelt(ctx, &qry->zone_cut);
+			return kr_error(EAGAIN);
+		}
+		/* No IPv4 nor IPv6, flag server as unusable. */
+		VERBOSE_MSG(qry, "=> unresolvable NS address, bailing out\n");
+		kr_zonecut_del_all(&qry->zone_cut, transport->ns_name);
+		return kr_error(EHOSTUNREACH);
+	}
+	/* Push new query to the resolution plan */
+	struct kr_query *next =
+		kr_rplan_push(rplan, qry, transport->ns_name, KNOT_CLASS_IN, next_type);
+	if (!next) {
+		return kr_error(ENOMEM);
+	}
+	next->flags.NONAUTH = true;
+
+	/* At the root level with no NS addresses, add SBELT subrequest. */
+	int ret = 0;
+	if (qry->zone_cut.name[0] == '\0') {
+		ret = kr_zonecut_set_sbelt(ctx, &next->zone_cut);
+		if (ret == 0) { /* Copy TA and key since it's the same cut to avoid lookup. */
+			kr_zonecut_copy_trust(&next->zone_cut, &qry->zone_cut);
+			kr_zonecut_set_sbelt(ctx, &qry->zone_cut); /* Add SBELT to parent in case query fails. */
+		}
+	} else {
+		next->flags.AWAIT_CUT = true;
+	}
+
+	if (ret == 0) {
+		if (next_type == KNOT_RRTYPE_AAAA) {
+			qry->flags.AWAIT_IPV6 = true;
+		} else {
+			qry->flags.AWAIT_IPV4 = true;
+		}	
+	}
+
+	return ret;
+}
+
+int kr_resolve_produce(struct kr_request *request, struct kr_transport **transport, knot_pkt_t *packet)
 {
 	struct kr_rplan *rplan = &request->rplan;
-	unsigned ns_election_iter = 0;
 
 	/* No query left for resolution */
 	if (kr_rplan_empty(rplan)) {
 		return KR_STATE_FAIL;
 	}
-	/* If we have deferred answers, resume them. */
+
 	struct kr_query *qry = array_tail(rplan->pending);
+
+	/* Initialize server selection */
+	if (!qry->server_selection.initialized) {
+		kr_server_selection_init(qry);
+	}
+
+	/* If we have deferred answers, resume them. */
 	if (qry->deferred != NULL) {
 		/* @todo: Refactoring validator, check trust chain before resuming. */
 		int state = 0;
@@ -1456,73 +1374,42 @@ int kr_resolve_produce(struct kr_request *request, struct sockaddr **dst, int *t
 		}
 	}
 
-ns_election:
-
-	if (unlikely(request->count_no_nsaddr >= KR_COUNT_NO_NSADDR_LIMIT)) {
-		VERBOSE_MSG(qry, "=> too many unresolvable NSs, bail out "
-				"(mitigation for NXNSAttack CVE-2020-12667)\n");
-		return KR_STATE_FAIL;
-	}
-	/* If the query has already selected a NS and is waiting for IPv4/IPv6 record,
-	 * elect best address only, otherwise elect a completely new NS.
-	 */
-	if(++ns_election_iter >= KR_ITER_LIMIT) {
-		VERBOSE_MSG(qry, "=> couldn't converge NS selection, bail out\n");
-		return KR_STATE_FAIL;
-	}
 
 	const struct kr_qflags qflg = qry->flags;
 	const bool retry = qflg.TCP || qflg.BADCOOKIE_AGAIN;
-	if (qflg.AWAIT_IPV4 || qflg.AWAIT_IPV6) {
-		kr_nsrep_elect_addr(qry, request->ctx);
-	} else if (qflg.FORWARD || qflg.STUB) {
-		kr_nsrep_sort(&qry->ns, request->ctx);
-		if (qry->ns.score > KR_NS_MAX_SCORE) {
-			/* At the moment all NS have bad reputation.
-			 * But there can be existing connections*/
-			VERBOSE_MSG(qry, "=> no valid NS left\n");
-			return KR_STATE_FAIL;
-		}
-	} else if (!qry->ns.name || !retry) { /* Keep NS when requerying/stub/badcookie. */
+	if (!qflg.FORWARD && !qflg.STUB && !retry) { /* Keep NS when requerying/stub/badcookie. */
 		/* Root DNSKEY must be fetched from the hints to avoid chicken and egg problem. */
 		if (qry->sname[0] == '\0' && qry->stype == KNOT_RRTYPE_DNSKEY) {
 			kr_zonecut_set_sbelt(request->ctx, &qry->zone_cut);
-			qry->flags.NO_THROTTLE = true; /* Pick even bad SBELT servers */
 		}
-		kr_nsrep_elect(qry, request->ctx);
-		if (qry->ns.score > KR_NS_MAX_SCORE) {
-			if (kr_zonecut_is_empty(&qry->zone_cut)) {
-				VERBOSE_MSG(qry, "=> no NS with an address\n");
-			} else {
-				VERBOSE_MSG(qry, "=> no valid NS left\n");
-			}
-			if (!qry->flags.NO_NS_FOUND) {
-				qry->flags.NO_NS_FOUND = true;
-			} else {
-				ITERATE_LAYERS(request, qry, reset);
-				kr_rplan_pop(rplan, qry);
-			}
+	}
+
+	qry->server_selection.choose_transport(qry, transport);
+
+	if (*transport == NULL) {
+		/* Properly signal to serve_stale module. */
+		if (qry->flags.NO_NS_FOUND) {
+			ITERATE_LAYERS(request, qry, reset);
+			kr_rplan_pop(rplan, qry);
+			return KR_STATE_FAIL;
+		} else {
+			/* FIXME: This is probably quite inefficient:
+			* we go through the whole qr_task_step loop just because of the serve_stale
+			* module which might not even be loaded. */
+			qry->flags.NO_NS_FOUND = true;
 			return KR_STATE_PRODUCE;
 		}
 	}
 
-	/* Resolve address records */
-	if (qry->ns.addr[0].ip.sa_family == AF_UNSPEC) {
-		int ret = ns_resolve_addr(qry, request);
-		if (ret != 0) {
-			qry->flags.AWAIT_IPV6 = false;
-			qry->flags.AWAIT_IPV4 = false;
-			qry->flags.TCP = false;
-			qry->ns.name = NULL;
-			goto ns_election; /* Must try different NS */
-		}
+	if ((*transport)->protocol == KR_TRANSPORT_RESOLVE_A || (*transport)->protocol == KR_TRANSPORT_RESOLVE_AAAA) {
+		uint16_t type = (*transport)->protocol == KR_TRANSPORT_RESOLVE_A ? KNOT_RRTYPE_A : KNOT_RRTYPE_AAAA;
+		ns_resolve_addr(qry, qry->request, *transport, type);
 		ITERATE_LAYERS(request, qry, reset);
 		return KR_STATE_PRODUCE;
 	}
 
-	/* Randomize query case (if not in safe mode or turned off) */
-	qry->secret = (qry->flags.SAFEMODE || qry->flags.NO_0X20)
-			? 0 : kr_rand_bytes(sizeof(qry->secret));
+	/* Randomize query case (if not in not turned off) */
+	qry->secret = qry->flags.NO_0X20 ? 0 : kr_rand_bytes(sizeof(qry->secret));
 	knot_dname_t *qname_raw = knot_pkt_qname(packet);
 	randomized_qname_case(qname_raw, qry->secret);
 
@@ -1531,8 +1418,6 @@ ns_election:
 	 * kr_resolve_checkout().
 	 */
 	qry->timestamp_mono = kr_now();
-	*dst = &qry->ns.addr[0].ip;
-	*type = (qry->flags.TCP) ? SOCK_STREAM : SOCK_DGRAM;
 	return request->state;
 }
 
@@ -1569,7 +1454,7 @@ static bool outbound_request_update_cookies(struct kr_request *req,
 #endif /* ENABLE_COOKIES */
 
 int kr_resolve_checkout(struct kr_request *request, const struct sockaddr *src,
-                        struct sockaddr *dst, int type, knot_pkt_t *packet)
+                        struct kr_transport *transport, knot_pkt_t *packet)
 {
 	/* @todo: Update documentation if this function becomes approved. */
 
@@ -1593,7 +1478,7 @@ int kr_resolve_checkout(struct kr_request *request, const struct sockaddr *src,
 		 * actual cookie. If we don't know the server address then we
 		 * also don't know the actual cookie size.
 		 */
-		if (!outbound_request_update_cookies(request, src, dst)) {
+		if (!outbound_request_update_cookies(request, src, &transport->address.ip)) {
 			return kr_error(EINVAL);
 		}
 	}
@@ -1610,8 +1495,20 @@ int kr_resolve_checkout(struct kr_request *request, const struct sockaddr *src,
 	/* Run the checkout layers and cancel on failure.
 	 * The checkout layer doesn't persist the state, so canceled subrequests
 	 * don't affect the resolution or rest of the processing. */
+	int type = -1;
+	switch(transport->protocol) {
+	case KR_TRANSPORT_UDP:
+		type = SOCK_DGRAM;
+		break;
+	case KR_TRANSPORT_TCP:
+	case KR_TRANSPORT_TLS:
+		type = SOCK_STREAM;
+		break;
+	default:
+		assert(0);
+	}
 	int state = request->state;
-	ITERATE_LAYERS(request, qry, checkout, packet, dst, type);
+	ITERATE_LAYERS(request, qry, checkout, packet, &transport->address.ip, type);
 	if (request->state & KR_STATE_FAIL) {
 		request->state = state; /* Restore */
 		return kr_error(ECANCELED);
@@ -1624,7 +1521,7 @@ int kr_resolve_checkout(struct kr_request *request, const struct sockaddr *src,
 	}
 
 	/* Write down OPT unless in safemode */
-	if (!(qry->flags.SAFEMODE)) {
+	if (!(qry->flags.NO_EDNS)) {
 		ret = edns_put(packet, true);
 		if (ret != 0) {
 			return kr_error(EINVAL);
@@ -1634,26 +1531,17 @@ int kr_resolve_checkout(struct kr_request *request, const struct sockaddr *src,
 	WITH_VERBOSE(qry) {
 
 	KR_DNAME_GET_STR(qname_str, knot_pkt_qname(packet));
+	KR_DNAME_GET_STR(ns_name, transport->ns_name);
 	KR_DNAME_GET_STR(zonecut_str, qry->zone_cut.name);
 	KR_RRTYPE_GET_STR(type_str, knot_pkt_qtype(packet));
+	const char *ns_str = kr_straddr(&transport->address.ip);
 
-	for (size_t i = 0; i < KR_NSREP_MAXADDR; ++i) {
-		struct sockaddr *addr = &qry->ns.addr[i].ip;
-		if (addr->sa_family == AF_UNSPEC) {
-			break;
-		}
-		if (!kr_inaddr_equal(dst, addr)) {
-			continue;
-		}
-		const char *ns_str = kr_straddr(addr);
-		VERBOSE_MSG(qry,
-			"=> id: '%05u' querying: '%s' score: %u zone cut: '%s' "
+	VERBOSE_MSG(qry,
+			"=> id: '%05u' querying: '%s'@'%s' zone cut: '%s' "
 			"qname: '%s' qtype: '%s' proto: '%s'\n",
-			qry->id, ns_str ? ns_str : "", qry->ns.score, zonecut_str,
+			qry->id, ns_name, ns_str ? ns_str : "", zonecut_str,
 			qname_str, type_str, (qry->flags.TCP) ? "tcp" : "udp");
-
-		break;
-	}}
+	}
 
 	return kr_ok();
 }
